package cyclops.collections.immutable;

import com.aol.cyclops2.data.collections.extensions.lazy.immutable.LazyPBagX;
import com.aol.cyclops2.data.collections.extensions.standard.LazyCollectionX;
import com.aol.cyclops2.types.recoverable.OnEmptySwitch;
import com.aol.cyclops2.types.foldable.To;
import com.aol.cyclops2.types.anyM.AnyMSeq;
import com.aol.cyclops2.types.foldable.ConvertableSequence.Conversion;
import cyclops.companion.Reducers;
import cyclops.collections.mutable.ListX;
import cyclops.control.Trampoline;
import cyclops.function.Fn3;
import cyclops.function.Fn4;
import cyclops.function.Monoid;
import cyclops.function.Reducer;
import cyclops.monads.AnyM;
import cyclops.monads.Witness.bagX;
import cyclops.stream.ReactiveSeq;
import cyclops.stream.Spouts;
import org.jooq.lambda.tuple.Tuple2;
import org.jooq.lambda.tuple.Tuple3;
import org.jooq.lambda.tuple.Tuple4;
import org.pcollections.HashTreePBag;
import org.pcollections.PBag;
import org.reactivestreams.Publisher;

import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.function.*;
import java.util.stream.Collector;
import java.util.stream.Stream;


/**
 *
 * A Lazy and Reactive Wrapper for Pesistent Bags (immutable collections like Sets that can contain duplicates).
 * BagX is lazy - chainging functional operations such as map / filter / flatMap only results in the collection
 * being traversed once. It is materialized and cached on first access (or via {@link BagX#materialize()}
 *
 * It is also possible to populate BagX asynchronously using reactive-streams publishers such as those created via the
 * {@link Spouts} class
 *
 *
 * @param <T>
 */
public interface BagX<T> extends To<BagX<T>>,PBag<T>, LazyCollectionX<T>, OnEmptySwitch<T, PBag<T>> {

    /**
     * Narrow a covariant BagX
     * 
     * <pre>
     * {@code 
     *  PBaagX<? extends Fruit> set = BagX.of(apple,bannana);
     *  BagX<Fruit> fruitSet = BagX.narrowK(set);
     * }
     * </pre>
     * 
     * @param bagX toNested narrowK generic type
     * @return BagX with narrowed type
     */
    public static <T> BagX<T> narrow(final BagX<? extends T> bagX) {
        return (BagX<T>) bagX;
    }
    
    /**
     * Create a BagX that contains the Integers between skip and take
     * 
     * @param start
     *            Number of range toNested skip from
     * @param end
     *            Number for range toNested take at
     * @return Range BagX
     */
    public static BagX<Integer> range(final int start, final int end) {
        return ReactiveSeq.range(start, end).to().bagX(Conversion.LAZY);
    }

    /**
     * Create a BagX that contains the Longs between skip and take
     * 
     * @param start
     *            Number of range toNested skip from
     * @param end
     *            Number for range toNested take at
     * @return Range BagX
     */
    public static BagX<Long> rangeLong(final long start, final long end) {
        return ReactiveSeq.rangeLong(start, end).to().bagX(Conversion.LAZY);

    }

    /**
     * Unfold a function into a BagX
     * 
     * <pre>
     * {@code 
     *  BagX.unfold(1,i->i<=6 ? Optional.of(Tuple.tuple(i,i+1)) : Optional.empty());
     * 
     * //(1,2,3,4,5)
     * 
     * }</code>
     * 
     * @param seed Initial value 
     * @param unfolder Iteratively applied function, terminated by an empty Optional
     * @return BagX generated by unfolder function
     */
    static <U, T> BagX<T> unfold(final U seed, final Function<? super U, Optional<Tuple2<T, U>>> unfolder) {
        return ReactiveSeq.unfold(seed, unfolder).to().bagX(Conversion.LAZY);
    }

    /**
     * Generate a BagX from the provided Supplier up toNested the provided limit number of times
     * 
     * @param limit Max number of elements toNested generate
     * @param s Supplier toNested generate BagX elements
     * @return BagX generated from the provided Supplier
     */
    public static <T> BagX<T> generate(final long limit, final Supplier<T> s) {

        return ReactiveSeq.generate(s)
                          .limit(limit).to().bagX(Conversion.LAZY);
    }
    /**
     * Generate a BagX from the provided value up toNested the provided limit number of times
     * 
     * @param limit Max number of elements toNested generate
     * @param s Value for BagX elements
     * @return BagX generated from the provided Supplier
     */
    public static <T> BagX<T> fill(final long limit, final T s) {

        return ReactiveSeq.fill(s)
                          .limit(limit).to().bagX(Conversion.LAZY);
    }
    @Override
    default BagX<T> materialize() {
        return (BagX<T>)LazyCollectionX.super.materialize();
    }

    /**
     * Create a BagX by iterative application of a function toNested an initial element up toNested the supplied limit number of times
     * 
     * @param limit Max number of elements toNested generate
     * @param seed Initial element
     * @param f Iteratively applied toNested each element toNested generate the next element
     * @return BagX generated by iterative application
     */
    public static <T> BagX<T> iterate(final long limit, final T seed, final UnaryOperator<T> f) {
        return ReactiveSeq.iterate(seed, f)
                          .limit(limit).to().bagX(Conversion.LAZY);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.data.collections.extensions.CollectionX#forEach4(java.util.function.Function, java.util.function.BiFunction, com.aol.cyclops2.util.function.TriFunction, com.aol.cyclops2.util.function.QuadFunction)
     */
    @Override
    default <R1, R2, R3, R> BagX<R> forEach4(Function<? super T, ? extends Iterable<R1>> stream1,
                                             BiFunction<? super T, ? super R1, ? extends Iterable<R2>> stream2,
                                             Fn3<? super T, ? super R1, ? super R2, ? extends Iterable<R3>> stream3,
                                             Fn4<? super T, ? super R1, ? super R2, ? super R3, ? extends R> yieldingFunction) {
        
        return (BagX)LazyCollectionX.super.forEach4(stream1, stream2, stream3, yieldingFunction);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.data.collections.extensions.CollectionX#forEach4(java.util.function.Function, java.util.function.BiFunction, com.aol.cyclops2.util.function.TriFunction, com.aol.cyclops2.util.function.QuadFunction, com.aol.cyclops2.util.function.QuadFunction)
     */
    @Override
    default <R1, R2, R3, R> BagX<R> forEach4(Function<? super T, ? extends Iterable<R1>> stream1,
                                             BiFunction<? super T, ? super R1, ? extends Iterable<R2>> stream2,
                                             Fn3<? super T, ? super R1, ? super R2, ? extends Iterable<R3>> stream3,
                                             Fn4<? super T, ? super R1, ? super R2, ? super R3, Boolean> filterFunction,
                                             Fn4<? super T, ? super R1, ? super R2, ? super R3, ? extends R> yieldingFunction) {
        
        return (BagX)LazyCollectionX.super.forEach4(stream1, stream2, stream3, filterFunction, yieldingFunction);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.data.collections.extensions.CollectionX#forEach3(java.util.function.Function, java.util.function.BiFunction, com.aol.cyclops2.util.function.TriFunction)
     */
    @Override
    default <R1, R2, R> BagX<R> forEach3(Function<? super T, ? extends Iterable<R1>> stream1,
                                         BiFunction<? super T, ? super R1, ? extends Iterable<R2>> stream2,
                                         Fn3<? super T, ? super R1, ? super R2, ? extends R> yieldingFunction) {
        
        return (BagX)LazyCollectionX.super.forEach3(stream1, stream2, yieldingFunction);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.data.collections.extensions.CollectionX#forEach3(java.util.function.Function, java.util.function.BiFunction, com.aol.cyclops2.util.function.TriFunction, com.aol.cyclops2.util.function.TriFunction)
     */
    @Override
    default <R1, R2, R> BagX<R> forEach3(Function<? super T, ? extends Iterable<R1>> stream1,
                                         BiFunction<? super T, ? super R1, ? extends Iterable<R2>> stream2,
                                         Fn3<? super T, ? super R1, ? super R2, Boolean> filterFunction,
                                         Fn3<? super T, ? super R1, ? super R2, ? extends R> yieldingFunction) {
        
        return (BagX)LazyCollectionX.super.forEach3(stream1, stream2, filterFunction, yieldingFunction);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.data.collections.extensions.CollectionX#forEach2(java.util.function.Function, java.util.function.BiFunction)
     */
    @Override
    default <R1, R> BagX<R> forEach2(Function<? super T, ? extends Iterable<R1>> stream1,
                                     BiFunction<? super T, ? super R1, ? extends R> yieldingFunction) {
        
        return (BagX)LazyCollectionX.super.forEach2(stream1, yieldingFunction);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.data.collections.extensions.CollectionX#forEach2(java.util.function.Function, java.util.function.BiFunction, java.util.function.BiFunction)
     */
    @Override
    default <R1, R> BagX<R> forEach2(Function<? super T, ? extends Iterable<R1>> stream1,
                                     BiFunction<? super T, ? super R1, Boolean> filterFunction,
                                     BiFunction<? super T, ? super R1, ? extends R> yieldingFunction) {
        
        return (BagX)LazyCollectionX.super.forEach2(stream1, filterFunction, yieldingFunction);
    }

    default AnyMSeq<bagX,T> anyM(){
        return AnyM.fromBagX(this);
    }
    
    @Override
    default BagX<T> take(final long num) {

        return limit(num);
    }
    @Override
    default BagX<T> drop(final long num) {

        return skip(num);
    }
    @Override
    default ReactiveSeq<T> stream() {

        return ReactiveSeq.fromIterable(this);
    }

    BagX<T> type(Reducer<? extends PBag<T>> reducer);
    
    /**
     *
     * <pre>
     * {@code
     *  import static cyclops.stream.ReactiveSeq.range;
     *
     *  BagX<Integer> bag = bagX(range(10,20));
     *
     * }
     * </pre>
     * @param stream To create BagX from
     * @param <T> BagX generated from Stream
     * @return
     */
    public static <T> BagX<T> bagX(ReactiveSeq<T> stream) {

        return new LazyPBagX<>(stream);
    }
    public static <T> BagX<T> of(final T... values) {
        return new LazyPBagX<>(
                               ReactiveSeq.of(values));
    }

    public static <T> BagX<T> empty() {
        return new LazyPBagX<>(
                               ReactiveSeq.empty());
    }

    public static <T> BagX<T> singleton(final T value) {
        //use concrete type for singleton as used in Reducers
        return new LazyPBagX<>(
                HashTreePBag.singleton(value));
    }

    /**
     * Construct a BagX from an Publisher
     * 
     * @param publisher
     *            toNested construct BagX from
     * @return BagX
     */
    public static <T> BagX<T> fromPublisher(final Publisher<? extends T> publisher) {
        return Spouts.from((Publisher<T>) publisher).to().bagX(Conversion.LAZY);
    }

    public static <T> BagX<T> fromIterable(final Iterable<T> iterable) {
        if (iterable instanceof BagX)
            return (BagX) iterable;
        if (iterable instanceof PBag)
            return new LazyPBagX<>(
                                   (PBag) iterable);


        return new LazyPBagX<>(null,
                                 ReactiveSeq.fromIterable(iterable),
                                 Reducers.toPBag());
    }



    default <T> BagX<T> fromStream(final ReactiveSeq<T> stream) {
        return new LazyPBagX<>(ReactiveSeq.fromStream(stream));
    }
    /**
     * coflatMap pattern, can be used toNested perform lazy reductions / collections / folds and other terminal operations
     * 
     * <pre>
     * {@code 
     *   
     *     BagX.of(1,2,3)
     *          .map(i->i*2)
     *          .coflatMap(s -> s.reduce(0,(a,b)->a+b))
     *      
     *      //BagX[12]
     * }
     * </pre>
     * 
     * 
     * @param fn mapping function
     * @return Transformed BagX
     */
    default <R> BagX<R> coflatMap(Function<? super BagX<T>, ? extends R> fn){
       return fn.andThen(r ->  this.<R>unit(r))
                .apply(this);
    }

    /**
    * Combine two adjacent elements in a BagX using the supplied BinaryOperator
    * This is a stateful grouping & reduction operation. The emitted of a combination may in turn be combined
    * with it's neighbor
    * <pre>
    * {@code 
    *  BagX.of(1,1,2,3)
                 .combine((a, b)->a.equals(b),Semigroups.intSum)
                 .listX()
                 
    *  //ListX(3,4) 
    * }</pre>
    * 
    * @param predicate Test toNested see if two neighbors should be joined
    * @param op Reducer toNested combine neighbors
    * @return Combined / Partially Reduced BagX
    */
    @Override
    default BagX<T> combine(final BiPredicate<? super T, ? super T> predicate, final BinaryOperator<T> op) {
        return (BagX<T>) LazyCollectionX.super.combine(predicate, op);
    }
    @Override
    default BagX<T> combine(final Monoid<T> op, final BiPredicate<? super T, ? super T> predicate) {
        return (BagX<T>)LazyCollectionX.super.combine(op,predicate);
    }



    /* (non-Javadoc)
     * @see com.aol.cyclops2.types.Pure#unit(java.lang.Object)
     */
    @Override
    default <R> BagX<R> unit(final R value) {
        return singleton(value);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.types.IterableFunctor#unitIterable(java.util.Iterator)
     */
    @Override
    default <R> BagX<R> unitIterator(final Iterator<R> it) {
        return fromIterable(() -> it);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.data.collections.extensions.persistent.LazyCollectionX#unit(java.util.Collection)
     */
    @Override
    default <R> BagX<R> unit(final Collection<R> col) {
        return fromIterable(col);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.data.collections.extensions.persistent.LazyCollectionX#emptyUnit()
     */
  //  @Override
    default <R> BagX<R> emptyUnit() {
        return empty();
    }



    /* (non-Javadoc)
     * @see com.aol.cyclops2.data.collections.extensions.persistent.LazyCollectionX#from(java.util.Collection)
     */
    @Override
    default <X> BagX<X> from(final Collection<X> col) {
        return fromIterable(col);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.data.collections.extensions.persistent.LazyCollectionX#monoid()
     */
 //   @Override
    default <T> Reducer<PBag<T>> monoid() {
        return Reducers.toPBag();
    }

    /* (non-Javadoc)
     * @see org.pcollections.PSet#plus(java.lang.Object)
     */
    @Override
    public BagX<T> plus(T e);

    /* (non-Javadoc)
     * @see org.pcollections.PSet#plusAll(java.util.Collection)
     */
    @Override
    public BagX<T> plusAll(Collection<? extends T> list);

    /* (non-Javadoc)
     * @see org.pcollections.PSet#minus(java.lang.Object)
     */
    @Override
    public BagX<T> minus(Object e);

    /* (non-Javadoc)
     * @see org.pcollections.PSet#minusAll(java.util.Collection)
     */
    @Override
    public BagX<T> minusAll(Collection<?> list);

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#reverse()
     */
    @Override
    default BagX<T> reverse() {
        return (BagX<T>) LazyCollectionX.super.reverse();
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#filter(java.util.function.Predicate)
     */
    @Override
    default BagX<T> filter(final Predicate<? super T> pred) {
        return (BagX<T>) LazyCollectionX.super.filter(pred);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#map(java.util.function.Function)
     */
    @Override
    default <R> BagX<R> map(final Function<? super T, ? extends R> mapper) {
        return (BagX<R>) LazyCollectionX.super.map(mapper);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#flatMap(java.util.function.Function)
     */
    @Override
    default <R> BagX<R> flatMap(final Function<? super T, ? extends Iterable<? extends R>> mapper) {
        return (BagX<R>) LazyCollectionX.super.flatMap(mapper);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#limit(long)
     */
    @Override
    default BagX<T> limit(final long num) {
        return (BagX<T>) LazyCollectionX.super.limit(num);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#skip(long)
     */
    @Override
    default BagX<T> skip(final long num) {
        return (BagX<T>) LazyCollectionX.super.skip(num);
    }

    @Override
    default BagX<T> takeRight(final int num) {
        return (BagX<T>) LazyCollectionX.super.takeRight(num);
    }

    @Override
    default BagX<T> dropRight(final int num) {
        return (BagX<T>) LazyCollectionX.super.dropRight(num);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#takeWhile(java.util.function.Predicate)
     */
    @Override
    default BagX<T> takeWhile(final Predicate<? super T> p) {
        return (BagX<T>) LazyCollectionX.super.takeWhile(p);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#dropWhile(java.util.function.Predicate)
     */
    @Override
    default BagX<T> dropWhile(final Predicate<? super T> p) {
        return (BagX<T>) LazyCollectionX.super.dropWhile(p);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#takeUntil(java.util.function.Predicate)
     */
    @Override
    default BagX<T> takeUntil(final Predicate<? super T> p) {
        return (BagX<T>) LazyCollectionX.super.takeUntil(p);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#dropUntil(java.util.function.Predicate)
     */
    @Override
    default BagX<T> dropUntil(final Predicate<? super T> p) {
        return (BagX<T>) LazyCollectionX.super.dropUntil(p);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#trampoline(java.util.function.Function)
     */
    @Override
    default <R> BagX<R> trampoline(final Function<? super T, ? extends Trampoline<? extends R>> mapper) {
        return (BagX<R>) LazyCollectionX.super.trampoline(mapper);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#slice(long, long)
     */
    @Override
    default BagX<T> slice(final long from, final long to) {
        return (BagX<T>) LazyCollectionX.super.slice(from, to);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#sorted(java.util.function.Function)
     */
    @Override
    default <U extends Comparable<? super U>> BagX<T> sorted(final Function<? super T, ? extends U> function) {
        return (BagX<T>) LazyCollectionX.super.sorted(function);
    }

    @Override
    default BagX<ListX<T>> grouped(final int groupSize) {
        return (BagX<ListX<T>>) LazyCollectionX.super.grouped(groupSize);
    }

    @Override
    default <K, A, D> BagX<Tuple2<K, D>> grouped(final Function<? super T, ? extends K> classifier, final Collector<? super T, A, D> downstream) {
        return (BagX) LazyCollectionX.super.grouped(classifier, downstream);
    }

    @Override
    default <K> BagX<Tuple2<K, ReactiveSeq<T>>> grouped(final Function<? super T, ? extends K> classifier) {
        return (BagX) LazyCollectionX.super.grouped(classifier);
    }

    @Override
    default <U> BagX<Tuple2<T, U>> zip(final Iterable<? extends U> other) {
        return (BagX) LazyCollectionX.super.zip(other);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#zip(java.lang.Iterable, java.util.function.BiFunction)
     */
    @Override
    default <U, R> BagX<R> zip(final Iterable<? extends U> other, final BiFunction<? super T, ? super U, ? extends R> zipper) {

        return (BagX<R>) LazyCollectionX.super.zip(other, zipper);
    }



    @Override
    default <U, R> BagX<R> zipS(final Stream<? extends U> other, final BiFunction<? super T, ? super U, ? extends R> zipper) {

        return (BagX<R>) LazyCollectionX.super.zipS(other, zipper);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#permutations()
     */
    @Override
    default BagX<ReactiveSeq<T>> permutations() {

        return (BagX<ReactiveSeq<T>>) LazyCollectionX.super.permutations();
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#combinations(int)
     */
    @Override
    default BagX<ReactiveSeq<T>> combinations(final int size) {

        return (BagX<ReactiveSeq<T>>) LazyCollectionX.super.combinations(size);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#combinations()
     */
    @Override
    default BagX<ReactiveSeq<T>> combinations() {

        return (BagX<ReactiveSeq<T>>) LazyCollectionX.super.combinations();
    }

    @Override
    default BagX<VectorX<T>> sliding(final int windowSize) {
        return (BagX<VectorX<T>>) LazyCollectionX.super.sliding(windowSize);
    }

    @Override
    default BagX<VectorX<T>> sliding(final int windowSize, final int increment) {
        return (BagX<VectorX<T>>) LazyCollectionX.super.sliding(windowSize, increment);
    }

    @Override
    default BagX<T> scanLeft(final Monoid<T> monoid) {
        return (BagX<T>) LazyCollectionX.super.scanLeft(monoid);
    }

    @Override
    default <U> BagX<U> scanLeft(final U seed, final BiFunction<? super U, ? super T, ? extends U> function) {
        return (BagX<U>) LazyCollectionX.super.scanLeft(seed, function);
    }

    @Override
    default BagX<T> scanRight(final Monoid<T> monoid) {
        return (BagX<T>) LazyCollectionX.super.scanRight(monoid);
    }

    @Override
    default <U> BagX<U> scanRight(final U identity, final BiFunction<? super T, ? super U, ? extends U> combiner) {
        return (BagX<U>) LazyCollectionX.super.scanRight(identity, combiner);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#plusInOrder(java.lang.Object)
     */
    @Override
    default BagX<T> plusInOrder(final T e) {

        return (BagX<T>) LazyCollectionX.super.plusInOrder(e);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#cycle(int)
     */
    @Override
    default BagX<T> cycle(final long times) {

        return (BagX<T>) LazyCollectionX.super.cycle(times);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#cycle(com.aol.cyclops2.sequence.Monoid, int)
     */
    @Override
    default BagX<T> cycle(final Monoid<T> m, final long times) {

        return (BagX<T>) LazyCollectionX.super.cycle(m, times);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#cycleWhile(java.util.function.Predicate)
     */
    @Override
    default BagX<T> cycleWhile(final Predicate<? super T> predicate) {

        return (BagX<T>) LazyCollectionX.super.cycleWhile(predicate);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#cycleUntil(java.util.function.Predicate)
     */
    @Override
    default BagX<T> cycleUntil(final Predicate<? super T> predicate) {

        return (BagX<T>) LazyCollectionX.super.cycleUntil(predicate);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#zip(java.util.reactiveStream.Stream)
     */
    @Override
    default <U> BagX<Tuple2<T, U>> zipS(final Stream<? extends U> other) {
        return (BagX) LazyCollectionX.super.zipS(other);
    }



    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#zip3(java.util.reactiveStream.Stream, java.util.reactiveStream.Stream)
     */
    @Override
    default <S, U> BagX<Tuple3<T, S, U>> zip3(final Iterable<? extends S> second, final Iterable<? extends U> third) {

        return (BagX) LazyCollectionX.super.zip3(second, third);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#zip4(java.util.reactiveStream.Stream, java.util.reactiveStream.Stream, java.util.reactiveStream.Stream)
     */
    @Override
    default <T2, T3, T4> BagX<Tuple4<T, T2, T3, T4>> zip4(final Iterable<? extends T2> second, final Iterable<? extends T3> third,
                                                          final Iterable<? extends T4> fourth) {

        return (BagX) LazyCollectionX.super.zip4(second, third, fourth);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#zipWithIndex()
     */
    @Override
    default BagX<Tuple2<T, Long>> zipWithIndex() {

        return (BagX<Tuple2<T, Long>>) LazyCollectionX.super.zipWithIndex();
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#distinct()
     */
    @Override
    default BagX<T> distinct() {

        return (BagX<T>) LazyCollectionX.super.distinct();
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#sorted()
     */
    @Override
    default BagX<T> sorted() {

        return (BagX<T>) LazyCollectionX.super.sorted();
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#sorted(java.util.Comparator)
     */
    @Override
    default BagX<T> sorted(final Comparator<? super T> c) {

        return (BagX<T>) LazyCollectionX.super.sorted(c);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#skipWhile(java.util.function.Predicate)
     */
    @Override
    default BagX<T> skipWhile(final Predicate<? super T> p) {

        return (BagX<T>) LazyCollectionX.super.skipWhile(p);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#skipUntil(java.util.function.Predicate)
     */
    @Override
    default BagX<T> skipUntil(final Predicate<? super T> p) {

        return (BagX<T>) LazyCollectionX.super.skipUntil(p);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#limitWhile(java.util.function.Predicate)
     */
    @Override
    default BagX<T> limitWhile(final Predicate<? super T> p) {

        return (BagX<T>) LazyCollectionX.super.limitWhile(p);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#limitUntil(java.util.function.Predicate)
     */
    @Override
    default BagX<T> limitUntil(final Predicate<? super T> p) {

        return (BagX<T>) LazyCollectionX.super.limitUntil(p);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#intersperse(java.lang.Object)
     */
    @Override
    default BagX<T> intersperse(final T value) {

        return (BagX<T>) LazyCollectionX.super.intersperse(value);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#shuffle()
     */
    @Override
    default BagX<T> shuffle() {

        return (BagX<T>) LazyCollectionX.super.shuffle();
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#skipLast(int)
     */
    @Override
    default BagX<T> skipLast(final int num) {

        return (BagX<T>) LazyCollectionX.super.skipLast(num);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#limitLast(int)
     */
    @Override
    default BagX<T> limitLast(final int num) {

        return (BagX<T>) LazyCollectionX.super.limitLast(num);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.types.recoverable.OnEmptySwitch#onEmptySwitch(java.util.function.Supplier)
     */
    @Override
    default BagX<T> onEmptySwitch(final Supplier<? extends PBag<T>> supplier) {
        if (isEmpty())
            return BagX.fromIterable(supplier.get());
        return this;
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#onEmpty(java.lang.Object)
     */
    @Override
    default BagX<T> onEmpty(final T value) {

        return (BagX<T>) LazyCollectionX.super.onEmpty(value);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#onEmptyGet(java.util.function.Supplier)
     */
    @Override
    default BagX<T> onEmptyGet(final Supplier<? extends T> supplier) {

        return (BagX<T>) LazyCollectionX.super.onEmptyGet(supplier);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#onEmptyThrow(java.util.function.Supplier)
     */
    @Override
    default <X extends Throwable> BagX<T> onEmptyThrow(final Supplier<? extends X> supplier) {

        return (BagX<T>) LazyCollectionX.super.onEmptyThrow(supplier);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#shuffle(java.util.Random)
     */
    @Override
    default BagX<T> shuffle(final Random random) {

        return (BagX<T>) LazyCollectionX.super.shuffle(random);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#ofType(java.lang.Class)
     */
    @Override
    default <U> BagX<U> ofType(final Class<? extends U> type) {

        return (BagX<U>) LazyCollectionX.super.ofType(type);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#filterNot(java.util.function.Predicate)
     */
    @Override
    default BagX<T> filterNot(final Predicate<? super T> fn) {

        return (BagX<T>) LazyCollectionX.super.filterNot(fn);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#notNull()
     */
    @Override
    default BagX<T> notNull() {

        return (BagX<T>) LazyCollectionX.super.notNull();
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#removeAll(java.util.reactiveStream.Stream)
     */
    @Override
    default BagX<T> removeAllS(final Stream<? extends T> stream) {

        return (BagX<T>) LazyCollectionX.super.removeAllS(stream);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#removeAll(java.lang.Iterable)
     */
    @Override
    default BagX<T> removeAllI(final Iterable<? extends T> it) {

        return (BagX<T>) LazyCollectionX.super.removeAllI(it);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#removeAll(java.lang.Object[])
     */
    @Override
    default BagX<T> removeAll(final T... values) {

        return (BagX<T>) LazyCollectionX.super.removeAll(values);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#retainAllI(java.lang.Iterable)
     */
    @Override
    default BagX<T> retainAllI(final Iterable<? extends T> it) {

        return (BagX<T>) LazyCollectionX.super.retainAllI(it);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#retainAllI(java.util.reactiveStream.Stream)
     */
    @Override
    default BagX<T> retainAllS(final Stream<? extends T> seq) {

        return (BagX<T>) LazyCollectionX.super.retainAllS(seq);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#retainAllI(java.lang.Object[])
     */
    @Override
    default BagX<T> retainAll(final T... values) {

        return (BagX<T>) LazyCollectionX.super.retainAll(values);
    }

    /* (non-Javadoc)
     * @see com.aol.cyclops2.collections.extensions.persistent.LazyCollectionX#cast(java.lang.Class)
     */
    @Override
    default <U> BagX<U> cast(final Class<? extends U> type) {

        return (BagX<U>) LazyCollectionX.super.cast(type);
    }



    @Override
    default <C extends Collection<? super T>> BagX<C> grouped(final int size, final Supplier<C> supplier) {

        return (BagX<C>) LazyCollectionX.super.grouped(size, supplier);
    }

    @Override
    default BagX<ListX<T>> groupedUntil(final Predicate<? super T> predicate) {

        return (BagX<ListX<T>>) LazyCollectionX.super.groupedUntil(predicate);
    }

    @Override
    default BagX<ListX<T>> groupedStatefullyUntil(final BiPredicate<ListX<? super T>, ? super T> predicate) {

        return (BagX<ListX<T>>) LazyCollectionX.super.groupedStatefullyUntil(predicate);
    }

    @Override
    default BagX<ListX<T>> groupedWhile(final Predicate<? super T> predicate) {

        return (BagX<ListX<T>>) LazyCollectionX.super.groupedWhile(predicate);
    }

    @Override
    default <C extends Collection<? super T>> BagX<C> groupedWhile(final Predicate<? super T> predicate, final Supplier<C> factory) {

        return (BagX<C>) LazyCollectionX.super.groupedWhile(predicate, factory);
    }

    @Override
    default <C extends Collection<? super T>> BagX<C> groupedUntil(final Predicate<? super T> predicate, final Supplier<C> factory) {

        return (BagX<C>) LazyCollectionX.super.groupedUntil(predicate, factory);
    }

    @Override
    default <R> BagX<R> retry(final Function<? super T, ? extends R> fn) {
        return (BagX<R>)LazyCollectionX.super.retry(fn);
    }

    @Override
    default <R> BagX<R> retry(final Function<? super T, ? extends R> fn, final int retries, final long delay, final TimeUnit timeUnit) {
        return (BagX<R>)LazyCollectionX.super.retry(fn);
    }

    @Override
    default <R> BagX<R> flatMapS(Function<? super T, ? extends Stream<? extends R>> fn) {
        return (BagX<R>)LazyCollectionX.super.flatMapS(fn);
    }

    @Override
    default <R> BagX<R> flatMapP(Function<? super T, ? extends Publisher<? extends R>> fn) {
        return (BagX<R>)LazyCollectionX.super.flatMapP(fn);
    }

    @Override
    default BagX<T> prependS(Stream<? extends T> stream) {
        return (BagX<T>)LazyCollectionX.super.prependS(stream);
    }

    @Override
    default BagX<T> append(T... values) {
        return (BagX<T>)LazyCollectionX.super.append(values);
    }

    @Override
    default BagX<T> append(T value) {
        return (BagX<T>)LazyCollectionX.super.append(value);
    }

    @Override
    default BagX<T> prepend(T value) {
        return (BagX<T>)LazyCollectionX.super.prepend(value);
    }

    @Override
    default BagX<T> prepend(T... values) {
        return (BagX<T>)LazyCollectionX.super.prepend(values);
    }

    @Override
    default BagX<T> insertAt(int pos, T... values) {
        return (BagX<T>)LazyCollectionX.super.insertAt(pos,values);
    }

    @Override
    default BagX<T> deleteBetween(int start, int end) {
        return (BagX<T>)LazyCollectionX.super.deleteBetween(start,end);
    }

    @Override
    default BagX<T> insertAtS(int pos, Stream<T> stream) {
        return (BagX<T>)LazyCollectionX.super.insertAtS(pos,stream);
    }

    @Override
    default BagX<T> recover(final Function<? super Throwable, ? extends T> fn) {
        return (BagX<T>)LazyCollectionX.super.recover(fn);
    }

    @Override
    default <EX extends Throwable> BagX<T> recover(Class<EX> exceptionClass, final Function<? super EX, ? extends T> fn) {
        return (BagX<T>)LazyCollectionX.super.recover(exceptionClass,fn);
    }

    @Override
    default BagX<T> plusLoop(int max, IntFunction<T> value) {
        return (BagX<T>)LazyCollectionX.super.plusLoop(max,value);
    }

    @Override
    default BagX<T> plusLoop(Supplier<Optional<T>> supplier) {
        return (BagX<T>)LazyCollectionX.super.plusLoop(supplier);
    }


    static <T> BagX<T> fromIterator(Iterator<T> iterator) {
        return fromIterable(()->iterator);
    }
}
